;;; Loads ECL's separate compiler and linker where it is not built-in.
;;; Remove the `nil' and adjust the path for ECL if necessary.
#+(and nil ecl (not osx))
(require "/usr/local/lib/ecl/cmp")

;;; Set up `readable-gensysms'.
(defvar *gensym-counter* 0)
(defun %gensym ()
  (intern (format nil "~~~A" (incf *gensym-counter*))))

(setf *compile-verbose* nil)
(setf *compile-print* nil)

; Unlocks CLISP's package `SOCKET' for sockets and `CL' for file-position
#+clisp
(eval-when (compile load eval)
  (when (>= (read-from-string (lisp-implementation-version) :end 4) 2.29)
    (pushnew :clisp-2.29 *features*)))
#+clisp-2.29 (setf (ext:package-lock '("SOCKET" "COMMON-LISP")) nil)

; Works around Corman Lisp weirdness and bugs
#+cormanlisp
(defun to-string (array)
  (let ((string (make-string (length array))))
    (dotimes (n (length array) string)
      (setf (char string n) (elt array n)))))

(defvar +proj-dir+ "~/slate/")
#+openmcl (setf +proj-dir+ (ccl::current-directory-name))

;;; Uncomment the following 2 lines to gain access to `subjective-dispatch'.
;;; They can work separately or together.
;(pushnew :subjective-slate *features*)
;(pushnew :layered-slate *features*)

;;; What files constitute the compiler. Use `cache' or `no-cache' depending on
;;; preference for more caching or smaller size.

(defconstant +files+ '("object" "no-cache" "compiler" "prims" "repl"))

;;; Compile the files.
(handler-bind ((warning #'muffle-warning)
	       (style-warning #'muffle-warning)
	       #+sbcl (sb-ext:compiler-note #'muffle-warning))
  (let ((cmp
	 (or #+ecl (lambda (name) (compile-file name :system-p t))
	     (lambda (name) (compile-file name)))))
    (mapc #'load +files+)
    (mapc cmp +files+)))

;;; Output the binary.

#+ecl
(progn
  (setf compiler::*cc-flags* "-g -I.")
  (c:build-program
   "slate" :lisp-files +files+ :ld-flags '()
   :epilogue-code "funcall(1,_intern(\"REPL\",user_package));"))
#+openmcl
(progn
  (setf ccl::*restore-lisp-functions* '(repl))
  (ccl::save-application "slate.image"))
#+cormanlisp
(save-application "slate" :console t :start-function #'repl :static t)
#+clisp (ext:saveinitmem "slate.mem" :quiet t :init-function #'repl)
