(define-object *no-role*
  (traits t nil))
(define-object *primitive-nil*
  (traits t nil))
(define-object *traits-method*
  (name nil "Method")
  (traits t nil)
  (cloneable t *primitive-nil*))
(define-object *primitive-method*
  (traits t *traits-method*)
  ("lineNumber" nil 0)
  ("fileName" nil "")
  (selector nil nil)
  (accessor nil nil)
  (arguments nil '())
  ("sourceCode" nil nil)
  (code nil nil))
(add-accessors *traits-method* '|cloneable|)
(add-accessors *traits-method* '|traits|)
(add-accessors *traits-method* '|name|)
(add-accessors *primitive-method* '|traits|)
(add-accessors *primitive-method* '|sourceCode|)
(add-accessors *primitive-method* '|lineNumber|)
(add-accessors *primitive-method* '|fileName|)
(add-accessors *primitive-nil* '|traits|)

(define-method "argumentNames" ((method *traits-method*))
  (coerce (get-slot method '|arguments|) 'vector))

(define-object *traits-root*
  (name nil "Root" t)
  (traits t nil t))
(define-object *primitive-root*
  (traits t *traits-root* t))

(define-object *traits-oddball*
  (name nil "Oddball" t)
  (traits t nil t)
  (root t *traits-root* t))
(define-object *primitive-oddball*
  (traits t *traits-oddball* t))
(set-slot *primitive-nil* '|traits| *traits-oddball*)
(set-slot *no-role* '|traits| *traits-oddball*)

(define-object *traits-derivable*
  (name nil "Derivable" t)
  (traits t nil t)
  (root t *traits-root* t))
(define-object *primitive-derivable*
  (traits t *traits-derivable* t))

(define-object *traits-cloneable*
  (name nil "Cloneable" t)
  (traits t nil t)
  (derivable t *traits-derivable* t))
(define-object *primitive-cloneable*
  (traits t *traits-cloneable* t))
(set-slot *traits-method* '|cloneable| *traits-cloneable*)

(define-object *traits-traits*
  (name nil "Traits" t)
  (root t *traits-root* t))
(define-object *primitive-traits*
  (name nil "_" t)
  (traits t *traits-traits* t))
(set-slot *traits-method* '|traits| *traits-traits*)
(set-slot *traits-root* '|traits| *traits-traits*)
(set-slot *traits-oddball* '|traits| *traits-traits*)
(set-slot *traits-derivable* '|traits| *traits-traits*)
(set-slot *traits-cloneable* '|traits| *traits-traits*)

(defmacro define-traits (name &rest slots)
  `(progn
     (define-object ,name (traits t *traits-traits* t) ,@slots)
     (define-method traits ((obj ,name))
       (if (eq obj ,name)
           *traits-traits*
           ,name))))

(defmacro define-bool-method (name arguments &body body)
  `(define-method ,name ,arguments
     (if (progn ,@body) *primitive-true* *primitive-false*)))

(defun add-to-namespace (object prototype)
  (add-slot object (invoke-method '|name| nil prototype) nil prototype t))

(define-traits *traits-lisp*
  (name nil "Lisp" t)
  (oddball t *traits-oddball* t))

(define-traits *traits-boolean*
  (name nil "Boolean" t)
  (oddball t *traits-oddball* t))
(define-object *primitive-true*
  (traits t *traits-boolean* t))
(define-object *primitive-false*
  (traits t *traits-boolean* t))

(define-traits *traits-number*
  (name nil "Number" t)
  (derivable t *traits-derivable* t))
(define-object *primitive-number*
  (traits t *traits-number* t))

(define-traits *traits-integer*
  (name nil "Integer" t)
  (number t *traits-number* t))

(define-traits *traits-float*
  (name nil "Float" t)
  (number t *traits-number* t))

(define-traits *traits-character*
  (name nil "Character" t)
  (oddball t *traits-oddball* t))

(define-traits *traits-symbol*
  (name nil "Symbol" t)
  (oddball t *traits-oddball* t))

(define-traits *traits-collection*
  (name nil "Collection" t)
  (cloneable t *traits-cloneable* t))
(define-object *primitive-collection*
  (traits t *traits-collection* t))

(define-traits *traits-array*
  (name nil "Array" t)
  (collection t *traits-collection* t))

(define-traits *traits-byte-array*
  (name nil "ByteArray" t)
  (collection t *traits-collection* t))

(define-traits *traits-word-array*
  (name nil "WordArray" t)
  (collection t *traits-collection* t))

(define-traits *traits-string*
  (name nil "String" t)
  (collection t *traits-collection*))

(define-traits *traits-file*
  (name nil "File" t)
  (oddball t *traits-oddball* t))
(define-object *primitive-file*
  (traits t *traits-file*))

#+clisp
(progn					; CLISP sockets.
(define-traits *traits-socket-client*
  (name nil "Socket" t)
  (oddball t *traits-oddball* t))
(define-object *primitive-socket-client*
  (traits t *traits-socket-client*))
(define-traits *traits-socket-server*
  (name nil "SocketServer" t)
  (oddball t *traits-oddball* t))
(define-object *primitive-socket-server*
  (traits t *traits-socket-server*))
)

(define-traits *traits-namespace*
  (name nil "Namespace" t)
  (oddball t *traits-oddball*))
(define-object *primitive-namespace*
  (traits t *traits-namespace*))

#+subjective-slate
(progn
  (define-traits *traits-subject*
    (name nil "Subject" t)
    (cloneable t *traits-cloneable*))

  (define-object *primitive-subject*
    (traits t *traits-subject*)
    ("parentSubject" t *primitive-nil* t)
    (focus t *primitive-nil* t)))

#+layered-slate
(progn
  (define-traits *traits-layer*
    (name nil "Layer" t)
    (cloneable t *traits-cloneable*))

  (define-object *primitive-layer*
    (traits t *traits-layer*)
    ("parentLayer" t *primitive-nil* t)
    (focus t *primitive-nil* t)))

(define-object *primitive-literal*
  (traits t *traits-oddball*))

(define-traits *traits-backtrace-frame*
  (name nil "BacktraceFrame" t)
  (cloneable t *traits-cloneable*))
  
(define-object *primitive-backtrace-frame*
  (traits t *traits-backtrace-frame*)
  (method nil *primitive-nil* t)
  (selector nil *primitive-nil* t)
  (arguments nil *primitive-nil* t)
  (optionals nil *primitive-nil* t))
  
(define-object *prototypes*
  (traits t *traits-namespace*)
  ("Traits" nil *primitive-traits* t)
  ("Root" nil *primitive-root* t)
  ("Cloneable" nil *primitive-cloneable* t)
  ("Oddball" nil *primitive-oddball* t)
  ("Derivable" nil *primitive-derivable* t)
  ("Namespace" nil *primitive-namespace* t)
  ("Nil" nil *primitive-nil* t)
  ("Boolean" nil *primitive-true* t)
  ("True" nil *primitive-true* t)
  ("False" nil *primitive-false* t)
  ("Number" nil *primitive-number* t)
  ("Integer" nil 0 t)
  ("Float" nil 0.0 t)
  ("ASCIICharacter" nil #+cormanlisp #\Nul #-cormanlisp #\Null t)
  ("Symbol" nil '|symbol| t)
  ("Collection" nil *primitive-collection* t)
  ("Array" nil (make-array 0 :initial-element *primitive-nil*) t)
  ("ByteArray" nil (make-array 0 :element-type '(unsigned-byte 8)) t)
  ("WordArray" nil (make-array 0 :element-type '(unsigned-byte 32)) t)
  ("String" nil "" t)
  ("File" nil *primitive-file* t)
  #+clisp ("SocketClient" nil *primitive-socket-client* t)
  #+clisp ("SocketServer" nil *primitive-socket-server* t)
  ("Lisp" nil (lambda () *primitive-nil*) t)
  ("BacktraceFrame" nil *primitive-backtrace-frame* t)
  #+subjective-slate ("Subject" nil *primitive-subject* t)
  #+layered-slate ("Layer" nil *primitive-layer* t)
  ("Method" nil *primitive-method* t))

(define-object *console-input*
  (traits t *traits-oddball* t))

(define-object *console-output*
  (traits t *traits-oddball* t))

(define-object *console*
  (traits t *traits-oddball* t)
  (reader nil *console-input* t)
  (writer nil *console-output* t))

(define-object *platform*
  (traits t *traits-oddball* t))

; Later replaced by an ExtensibleSequence.
(defvar *slate-modules*
  (make-array 12 :initial-element nil :fill-pointer 0 :adjustable t))

; Later replaced by an ExtensibleSequence.
(defvar *slate-features*
  (make-array 12 :initial-element nil :fill-pointer 0 :adjustable t))

#+clisp
(progn
(define-object *shell*
  (traits t *traits-oddball*)
  (name nil "Shell"))
(define-object *environment*
  (traits t *traits-oddball*)
  (name nil "Environment"))
)

(define-object *globals*
  (traits t *traits-namespace*)
  ("NoRole" nil *no-role* t)
  #+clisp ("Shell" nil *shell* t)
  #+clisp ("Environment" nil *environment* t)
  ("Console" nil *console* t)
  ("ConsoleInput" nil *console-input* t)
  ("ConsoleOutput" nil *console-output* t)
  ("modules" nil *slate-modules* t)
  ("features" nil *slate-features* t)
  ("currentModule" nil *primitive-nil* t)
  ("platform" nil *platform* t))

(define-object *lobby*
  (traits t *traits-namespace*)
  (lobby nil *primitive-nil* t)
  (prototypes t *prototypes* t)
  (globals t *globals* t))

(set-slot *lobby* '|lobby| *lobby*)

#+subjective-slate
(progn
  (defvar *subject* nil)

  (define-method "withoutSubject" ((method *traits-method*))
    (let ((old-subject *subject*))
      (prog2
        (setf *subject* nil)
        (apply-method method)
        (setf *subject* old-subject))))
        
  (define-method "seenFrom:" ((method *traits-method*) focus)
    (let ((new-subject (clone-object *primitive-subject*))
          (old-subject *subject*))
      (set-slot new-subject '|focus| focus)
      (set-slot new-subject '|parentSubject| old-subject)
      (prog2
        (setf *subject* new-subject)
        (apply-method method)
        (setf *subject* old-subject)))))

#+layered-slate
(progn
  (defvar *layer* nil)

  (define-method "withoutLayers" ((method *traits-method*))
    (let ((old-layer *layer*))
      (prog2
        (setf *layer* nil)
        (apply-method method)
        (setf *layer* old-layer))))

  (define-method "layering:" ((method *traits-method*) focus)
    (let ((new-layer (clone-object *primitive-layer*))
          (old-layer *layer*))
      (set-slot new-layer '|focus| focus)
      (set-slot new-layer '|parentLayer| old-layer)
      (prog2
        (setf *layer* new-layer)
        (apply-method method)
        (setf *layer* old-layer)))))

(unless (fboundp 'exit)
  (setf (fdefinition 'exit) #'quit))
(define-method "exit" ((lobby *lobby*))
  (exit))
(define-method "quit" ((lobby *lobby*))
  (exit))

#+clisp
(define-method "saveImageNamed:" ((lobby *lobby*) filename)
  (if (ext:saveinitmem filename :quiet t :init-function #'repl)
      *primitive-true* *primitive-false*))

#+clisp
(progn					; Shell-related methods.
(define-method "runProgram:withArgs:" ((shell *shell*) prog-name args)
  (apply 'ext:execute `(,prog-name ,@(coerce args 'list))))
(define-method "execute:" ((shell *shell*) command-string)
  (ext:shell command-string))
(define-method "enter" ((shell *shell*))
  (ext:shell))
(define-method "run:withArgs:inputFrom:outputTo:"
  ((shell *shell*) command-string args in out)
  (apply 'ext:run-shell-command
	 `(,command-string ,@(coerce args 'list) :input ,in :output ,out)))
(define-method "at:" ((environment *environment*) env-string)
  (ext:getenv env-string))
(define-method "at:put:" ((environment *environment*) env-string value-string)
  (setf (ext:getenv env-string) value-string))
(define-method "keys" ((environment *environment*))
  (coerce
   (loop for pair in (ext:getenv)
     collect (car pair))
   'vector))
(define-method "values" ((environment *environment*))
  (coerce
   (loop for pair in (ext:getenv)
     collect (cdr pair))
   'vector))
(define-traits *traits-pipe*
  (name nil "Pipe")
  (traits t *traits-oddball*))
(define-object *pipe-primitive*
  (traits t *traits-pipe*)
  (command nil ""))
(define-method "newForCommand:" ((pipe *traits-pipe*) command-string)
  ())
)

; Only here for bootstrapping the libraries.
(define-method "requires:" ((lobby *lobby*) (features *traits-array*))
  (loop for required across features
    when (notany (lambda (present) (eq required present)) *slate-features*)
    do (format t "Feature ~A missing for module ~A.~%"
	       required (get-slot *globals* '|currentModule|)))
  *primitive-nil*)

; Only here for bootstrapping the libraries.
(define-method "provides:" ((lobby *lobby*) (features *traits-array*))
  (loop for provided across features
    unless (some (lambda (present) (eq provided present)) *slate-features*)
    do (vector-push-extend provided *slate-features*))
  *slate-features*)

; Only here for bootstrapping the libraries.
(define-method "enterModule:" ((lobby *lobby*) name)
  (vector-push-extend name *slate-modules*)
  (setf (get-slot *globals* '|currentModule|) name))

(define-bool-method "debug" ((lobby *lobby*))
  *slate-debug*)

(define-bool-method "debug:" ((lobby *lobby*) (flag *traits-boolean*))
  (setf *slate-debug* flag))

(define-method "bits" ((float *traits-float*))
  (multiple-value-bind (significand exponent sign) (integer-decode-float float)
    (loop until (or (logbitp 0 significand) (= exponent 127))
      do (setf significand (ash significand -1))
         (incf exponent))
    (logior (if (< sign 0) (ash 1 31) 0)
            (ash significand 23)
            (+ exponent 127))))

(define-method "as:" ((number *traits-number*) (string *traits-string*))
  (format nil "~A" number))

(define-method "as:" ((string *traits-string*) (number *traits-number*))
  (values (read-from-string string)))

(define-method "as:" ((string *traits-string*) (integer *traits-integer*))
  (values (parse-integer string)))

(define-method "as:" ((character *traits-character*) (integer *traits-integer*))
  (char-code character))

(define-method "as:" ((integer *traits-integer*) (character *traits-character*))
  (code-char integer))

(define-method "as:" ((selector *traits-symbol*) (string *traits-string*))
  (symbol-name selector))

(define-method "as:" ((string *traits-string*) (selector *traits-symbol*))
  (intern string))

(define-method "next" ((stream *console-input*))
  (read-char))

(define-method "flush" ((stream *console-input*))
  (finish-output stream))

(define-method "nextPut:" ((stream *console-output*) (character *traits-character*))
  (write-char character))

(define-method "flush" ((stream *console-output*))
  (finish-output stream))

(define-method clone ((oddball *traits-oddball*))
  (if (eq oddball *primitive-oddball*)
      (clone-object *primitive-oddball*)
      oddball))

(define-method clone ((namespace *traits-namespace*))
  (if (eq namespace *primitive-namespace*)
      (clone-object *primitive-namespace*)
      namespace))

; Only here for bootstrapping the libraries.
(define-method derive ((derivable *traits-derivable*))
  (let ((object (invoke-method '|clone| nil derivable))
        (traits (clone-object *primitive-traits*)))
    (add-slot traits '|parent| t (invoke-method '|traits| nil object) t)
    (set-slot object '|traits| traits)
    object))

; Only here for bootstrapping the libraries.
(define-method "deriveWith:" ((derivable *traits-derivable*) array)
  (let ((traits (clone-object *primitive-traits*)))
    (loop for parent in (nreverse (cons derivable (coerce array 'list)))
          for index from 0
          with object
      do
        (if object
            (invoke-method '|addSlotsFrom:| nil object parent)
            (setf object (invoke-method '|clone| nil parent)))
        (add-slot traits (intern (format nil "parent~A" index)) t (invoke-method '|traits| nil parent) t)
      finally 
        (set-slot object '|traits| traits)
        (return object))))

; Only here for bootstrapping the libraries.
(define-method "allButFirst" ((array *traits-array*))
  (subseq array 1))

(define-method "addPrototype:derivedFrom:" ((obj *traits-root*) name array)
  (let* ((proto (if (eq (length array) 1)
		    (invoke-method '|derive| nil (svref array 0))
		  (invoke-method '|deriveWith:| nil
				 (svref array 0) (subseq array 1))))
	 (traits (get-slot proto '|traits|)))
    (add-slot obj name nil proto t)
    (multiple-value-bind (value exists) (get-slot traits '|name|)
      (if exists
	  (set-slot traits '|name| (symbol-name name))
	(add-slot traits '|name| nil (symbol-name name))))
    proto))

(define-bool-method "==" ((x *traits-root*) (y *traits-root*))
  (eq x y))

(define-method "asMethod:on:" ((method *traits-method*) (selector *traits-symbol*) array)
  (let ((method (clone-object method)))
    (set-slot method '|selector| selector)
    (apply #'add-roles selector method (coerce array 'list))))

; Only here for bootstrapping the libraries.
(define-method "addSlotsFrom:" ((x *traits-root*) (y *traits-root*))
  (loop for slot-name in (wm-map-slot-names (wm-object-map y))
        for slot = (gethash slot-name (wm-map-slots (wm-object-map y)))
        for contents = (svref (wm-object-slots y) (wm-slot-index slot))
    do
      (add-slot x slot-name
        (wm-slot-delegates slot)
        contents
        (wm-slot-accessors slot)))
  x) 

(define-method "addRolesFrom:" ((x *traits-root*) (y *traits-root*))
  (let ((map 
          (if (wm-object-p y) 
              (wm-object-map y)
              (wm-object-map (gethash y *literal-objects*)))))
    (maphash
      #'(lambda (name slot)
          (loop for role in (wm-slot-roles slot)
                for method = (wm-role-method role)
            when (eq (get-slot method '|accessor|) *primitive-nil*)
              do (add-role x name (wm-role-positions role) (wm-role-arguments role) method)))
     (wm-map-slots map))
    x))

(define-method "addSlot:valued:" ((object *traits-root*) name value)
  (add-slot object (if (symbolp name) name (intern name)) nil value t)
  object)

(define-method "addDelegate:valued:" ((object *traits-root*) name value)
  (add-slot object (if (symbolp name) name (intern name)) t value t)
  object)

(define-method "addImmutableSlot:valued:" ((object *traits-root*) name value)
  (add-slot object (if (symbolp name) name (intern name)) nil value :immutable)
  object)

(define-method "addImmutableDelegate:valued:" ((object *traits-root*) name value)
  (add-slot object (if (symbolp name) name (intern name)) t value :immutable)
  object)

(define-method "removeSlotNamed:" ((object *traits-root*) name)
  (remove-slot object (if (symbolp name) name (intern name))))

(define-method "slotNames" ((object *traits-root*))
  (unless (wm-object-p object)
    (setf object (gethash object *literal-objects*))
    (unless object (throw 'top-level #(|traits|))))
  (coerce (wm-map-slot-names (wm-object-map object)) 'vector))

(define-method "delegateNames" ((object *traits-root*))
  (unless (wm-object-p object)
    (setf object (gethash object *literal-objects*))
    (unless object (throw 'top-level #(|traits|))))
  (coerce 
   (loop for slot-name in (wm-map-slot-names (wm-object-map object))
     for slot = (gethash slot-name (wm-map-slots (wm-object-map object)))
     when (wm-slot-delegates slot)
     collect slot-name)
   'vector))

(define-method clone ((object *traits-root*))
  (if (wm-object-p object)
      (clone-object object)
      (clone-object
        (or (gethash object *literal-objects*)
            (let ((representation (make-wm-object)))
              (add-slot representation '|traits| t
                (traits-for-prim object) t)
              representation)))))

(define-method clone ((array *traits-byte-array*))
  (let ((new-array (copy-seq array))
        (representation (gethash array *literal-objects*)))
    (setf (gethash new-array *literal-objects*)
      (if representation
          (clone-object representation)
          (let ((representation (make-wm-object)))
            (add-slot representation '|traits| t *traits-byte-array*)
            representation)))
    new-array))

(define-method clone ((array *traits-word-array*))
  (let ((new-array (copy-seq array))
        (representation (gethash array *literal-objects*)))
    (setf (gethash new-array *literal-objects*)
      (if representation
          (clone-object representation)
          (let ((representation (make-wm-object)))
            (add-slot representation '|traits| t *traits-word-array*)
            representation)))
    new-array))

(define-method clone ((array *traits-array*))
  (let ((new-array (copy-seq array))
        (representation (gethash array *literal-objects*)))
    (setf (gethash new-array *literal-objects*)
      (if representation
          (clone-object representation)
          (let ((representation (make-wm-object)))
            (add-slot representation '|traits| t *traits-array*)
            representation)))
    new-array))

(define-method clone ((array *traits-string*))
  (let ((new-array (copy-seq array))
        (representation (gethash array *literal-objects*)))
    (setf (gethash new-array *literal-objects*)
      (if representation
          (clone-object representation)
          (let ((representation (make-wm-object)))
            (add-slot representation '|traits| t *traits-string*)
            representation)))
    new-array))

(define-method "newSize:" ((array *traits-byte-array*) (size *traits-integer*))
  (let ((representation (gethash array *literal-objects*))
        (new-array (make-array size :element-type '(unsigned-byte 8) :initial-element 0)))
    (when representation
      (setf (gethash new-array *literal-objects*) (clone-object representation)))
    new-array))

(define-method "newSize:" ((array *traits-word-array*) (size *traits-integer*))
  (let ((representation (gethash array *literal-objects*))
        (new-array (make-array size :element-type '(unsigned-byte 32) :initial-element 0)))
    (when representation
      (setf (gethash new-array *literal-objects*) (clone-object representation)))
    new-array))

(define-method "newSize:" ((array *traits-array*) (size *traits-integer*))
  (let ((representation (gethash array *literal-objects*))
        (new-array (make-array size :initial-element *primitive-nil*)))
    (when representation
      (setf (gethash new-array *literal-objects*) (clone-object representation)))
    new-array))

(define-method "newSize:" ((array *traits-string*) (size *traits-integer*))
  (let ((representation (gethash array *literal-objects*))
        (new-array (make-array size :element-type 'character :initial-element 
                #+cormanlisp #\Nul #-cormanlisp #\Null)))
    (when representation
      (setf (gethash new-array *literal-objects*) (clone-object representation)))
    new-array))

(define-method "atSlotIndex:" ((object *traits-root*) (index *traits-integer*))
  (unless (wm-object-p object)
    (setf object (gethash object *literal-objects*)))
  (svref (wm-object-slots object) index))

(define-method "atSlotIndex:put:" ((object *traits-root*) (index *traits-integer*) value)
  (unless (wm-object-p object)
    (setf object (gethash object *literal-objects*)))
  (setf (svref (wm-object-slots object) index) value))

(define-method "atSlotNamed:" ((object *traits-root*) (slot *traits-symbol*))
  (if (and (not (wm-object-p object))
           (eq slot '|traits|))
      (invoke-method '|traits| nil object)
      (get-slot object slot)))

(define-method "atSlotNamed:put:" ((object *traits-root*) (slot *traits-symbol*) value)
  (set-slot object slot value))

(define-method "at:" ((array *traits-byte-array*) (index *traits-integer*))
  (aref array index))

(define-method "at:put:" ((array *traits-byte-array*) (index *traits-integer*) value)
  (setf (aref array index) value))

(define-method size ((array *traits-byte-array*))
  (length array))

(define-method "at:" ((array *traits-word-array*) (index *traits-integer*))
  (aref array index))

(define-method "at:put:" ((array *traits-word-array*) (index *traits-integer*) value)
  (setf (aref array index) value))

(define-method size ((array *traits-word-array*))
  (length array))

(define-method "at:" ((array *traits-array*) (index *traits-integer*))
  (aref array index))

(define-method "at:put:" ((array *traits-array*) (index *traits-integer*) value)
  (setf (aref array index) value))

(define-bool-method "includes:" ((array *traits-array*) obj)
  (find obj array))

(define-method size ((array *traits-array*))
  (length array))

(define-method "at:" ((string *traits-string*) (index *traits-integer*))
  (char string index))

(define-method "at:put:" ((string *traits-string*) (index *traits-integer*) value)
  (setf (char string index) value))

(define-method size ((string *traits-string*))
  (length string))

; This anticipates the loading of the Streams library.
(define-method "printOn:" ((object *traits-root*) (stream *traits-root*))
  (invoke-method '|nextPutAll:| nil stream
		 (invoke-method '|print| nil object)))

(define-method "printOn:" ((object *traits-root*) (stream *console-output*))
  (format t "~A" (invoke-method '|print| nil object)))

(define-method print ((object *traits-traits*))
  (unless (and (wm-object-p object)
               (or (eq object *traits-traits*)
                   (eq (get-slot object '|traits|) *traits-traits*)))
    (return-from print (invoke-method '|print| currentMethod object)))
  (format nil "@~A" (or (get-slot object '|name|) "_")))

(define-method print ((object *traits-symbol*))
  (when (wm-object-p object)
    (return-from print
      (invoke-method '|print| currentMethod object)))
  (format nil "#~A" object))

(define-method print ((object *traits-character*))
  (when (wm-object-p object)
    (return-from print
      (invoke-method '|print| currentMethod object)))
  (format nil "$~A"
    (case object
      (#\Newline "\\n")
      (#\Tab "\\t")
      (#\Return "\\r")
      (#\Backspace "\\b")
      (#\Space "\\s")
      (#\^[ "\\e")
      ((or #+cormanlisp #\Nul #\Null) "\\0")
      (t
        (case (char-code object)
          (7 "\\a")
          (11 "\\v")
          (12 "\\f")
          (t object))))))

(define-method print ((object *traits-root*))
  (let ((string "<"))
    (when (get-slot object '|traits|)
      (setf string
	    (format nil "~A@~A:" string
		    (or (get-slot (get-slot object '|traits|) '|name|) "_"))))
    (setf string
	  (if (< (length (wm-map-slot-names (wm-object-map object))) 20)
	      (format nil "~A ~{ ~A~}>"
		      string (wm-map-slot-names (wm-object-map object)))
	    (format nil "~A ...>" string)))
    string))

(define-method print ((object *primitive-nil*))
  "Nil")

(define-method print ((object *primitive-true*))
  "True")

(define-method print ((object *primitive-false*))
  "False")

(define-method print ((object *no-role*))
  "NoRole")

(define-method print ((file *traits-file*))
  (when (wm-object-p file)
    (return-from print
      (invoke-method '|print| currentMethod file)))
  (format nil "<File ~A>" (namestring file)))

(define-method print ((object *traits-lisp*))
  (when (wm-object-p object)
    (return-from print
      (invoke-method '|print| currentMethod object)))
  (format nil "<Lambda ~A>" (function-lambda-expression object)))

(define-method print ((object *traits-number*))
  (when (wm-object-p object)
    (return-from print
      (invoke-method '|print| currentMethod object)))
  (if (wm-object-p object)
      "<@Number>"
      (format nil "~A" object)))

(define-method print ((array *traits-array*))
  (when (wm-object-p array)
    (return-from print
      (invoke-method '|print| currentMethod array))) 
  (let ((string "{"))
    (setf string
      (apply #'concatenate 'string string
        (loop for index below (1- (length array))
          collect (invoke-method '|print| nil (aref array index))
          collect (format nil ". "))))
    (when (> (length array) 0)
      (setf string
        (concatenate 'string string
          (invoke-method '|print| nil (aref array (1- (length array)))))))
    (setf string (concatenate 'string string "}"))
    string))

(define-method print ((array *traits-byte-array*))
  (when (wm-object-p array)
    (return-from print
      (invoke-method '|print| currentMethod array)))
  (let ((string "{"))
    (setf string
      (apply #'concatenate 'string string
        (loop for index below (1- (length array))
          collect (format nil "~A. " (aref array index)))))
    (when (> (length array) 0)
      (setf string
        (concatenate 'string string (format nil "~A" (aref array (1- (length array)))))))
    (setf string (concatenate 'string string "}"))
    string))

(define-method print ((array *traits-word-array*))
  (when (wm-object-p array)
    (return-from print
      (invoke-method '|print| currentMethod array)))
  (let ((string "{"))
    (setf string
      (apply #'concatenate 'string string
        (loop for index below (1- (length array))
          collect (format nil "~A. " (aref array index)))))
    (when (> (length array) 0)
      (setf string
        (concatenate 'string string (format nil "~A" (aref array (1- (length array)))))))
    (setf string (concatenate 'string string "}"))
    string))

(define-method print ((string *traits-string*))
  (when (wm-object-p string)
    (return-from print
      (invoke-method '|print| currentMethod string))) 
  (format nil "'~A'" string))

(define-method print ((method *traits-method*))
  (when (eq method *traits-method*)
    (return-from print
      (invoke-method '|print| currentMethod method))) 
  (format nil "[~A]" (if (eq (get-slot method '|selector|) *primitive-nil*) "" (get-slot method '|selector|))))

(define-method loop ((method *traits-method*))
  (loop (apply-method method)))

(define-method value ((method *traits-method*))
  (apply-method method))

(define-method "value:" ((method *traits-method*) x)
  (apply-method method x))

(define-method "value:value:" ((method *traits-method*) x y)
  (apply-method method x y))

(define-method "value:value:value:" ((method *traits-method*) x y z)
  (apply-method method x y z))

(define-method "values:" ((method *traits-method*) arguments &optionals optionals)
  (apply (get-slot method '|code|) method
    (if optionals
        (nconc (coerce arguments 'list) (list optionals))
        (coerce arguments 'list))))

(define-method "type" ((platform *platform*))
  (machine-type))

(define-method "version" ((platform *platform*))
  (machine-version))

(define-method "OS" ((platform *platform*))
  (software-type))

(define-method "OSversion" ((platform *platform*))
  (software-version))

(define-method "newNamed:" ((file *traits-file*) (filename *traits-string*))
  (open (parse-namestring filename) :direction :io :if-exists :overwrite))

(define-method "newForInputNamed:" ((file *traits-file*) (filename *traits-string*))
  (or (open (parse-namestring filename) :direction :input :if-does-not-exist nil) *primitive-nil*))

(define-method "newForOutputNamed:" ((file *traits-file*) (filename *traits-string*))
  (open (parse-namestring filename) :direction :output :if-exists :append))

(define-method "newForNewNamed:" ((file *traits-file*) (filename *traits-string*))
  (open (parse-namestring filename)
	:direction :append :if-exists :overwrite :if-does-not-exist :create))

(define-bool-method "close" ((file *traits-file*))
  (close file))

(define-method "fullName" ((file *traits-file*))
  (namestring (truename file)))

(define-method "renameTo:" ((file *traits-file*) (newname *traits-string*))
  (multiple-value-bind (new-name old-path new-path) (rename-file file newname)
    (namestring new-path)))

(define-bool-method "delete" ((file *traits-file*))
  (delete-file file))

(define-method "size" ((file *traits-file*))
  (file-length file))

; The following methods are intended to write to Sequences of some kind, so
; that Arrays, Strings, and ByteArrays may be used equally.
(define-method "read:from:startingAt:into:" ((file *traits-file*)
					     n handle start s)
  (read-sequence s file :start start))

(define-method "write:to:startingAt:from:" ((file *traits-file*)
					    n handle start s)
  (write-sequence s file :start start))

(define-method "position" ((file *traits-file*))
  (file-position file))

#-cormanlisp
(define-bool-method "position:" ((file *traits-file*) pos)
  (file-position file pos))

(define-method "size" ((file *traits-file*))
  (file-length file))

(define-method "filename" ((file *traits-file*))
  (namestring file))

(define-bool-method "debugMode" ((lobby *lobby*))
  *slate-debuggable*)

(define-method "debugMode:" ((lobby *lobby*) (flag *traits-boolean*))
  (setf *slate-debuggable* (eq flag *primitive-true*))
  flag)

(define-method "backtrace" ((lobby *lobby*))
  (coerce
    (loop for (method selector arguments optionals) in *slate-backtrace*
          for backtrace = (clone-object *primitive-backtrace-frame*)
      do (set-slot backtrace '|method| method)
         (set-slot backtrace '|selector| selector)
         (set-slot backtrace '|arguments| arguments)
         (set-slot backtrace '|optionals| (or optionals *primitive-nil*))
      collect backtrace)
      'simple-vector))

(define-method "ensure:" ((method *traits-method*) (final *traits-method*))
  (unwind-protect (apply-method method)
    (apply-method final)))

(define-method "initHook" ((lobby *lobby*))
  *primitive-nil*)

(define-method "interpretHook" ((method *traits-method*))
  (or (apply-method method) *primitive-nil*))

(define-bool-method "compileMode" ((lobby *lobby*))
  *slate-compile*)

(define-method "compileMode:" ((lobby *lobby*) (flag *traits-boolean*))
  (setf *slate-compile* (eq flag *primitive-true*))
  flag)

(define-method "compileFileNamed:" ((lobby *lobby*) (filename *traits-string*))
  (let ((out-name
	 (concatenate 'string
		      (subseq filename 0 (search ".slate" filename)) ".lisp")))
    (compile-slate-to-lisp filename out-name)
    (compile-file out-name :verbose nil :print nil)
    (delete-file out-name))
  *primitive-true*)

(define-method "fileIn:" ((file *traits-file*)
			  (string *traits-string*))
  (when (not (probe-file string))
    (format t "File '~A' not found. Nothing loaded." string)
    (throw 'top-level nil))
  (let ((fasl-name
	 (compile-file-pathname (subseq string 0 (search ".slate" string)))))
    (when (and (probe-file fasl-name) (not (search ".test" string)))
      (format t "~&Loading '~A'.~%" fasl-name)
      (load fasl-name :verbose nil :print nil)
      (return *primitive-nil*)))
  (invoke-method '|enterModule:| nil *lobby* string)
  (with-open-file (file string)
    (with-file-name (file-name string)
      (with-line-number (line-number 0)
	(let ((token-stream (make-token-stream :stream file)))
#-ecl
	  (loop for code = (parse-statement token-stream :eof)
	    with result = *primitive-nil*
	    do (setf result (compile-slate code))
	    if (eq (peek-token token-stream) :end-statement)
	      do (get-token token-stream)
	    else
	      if (eq (peek-token token-stream) :eof)
	      do (return result))
#+ecl
	  (funcall
	   (if *slate-compile* #'compile-slate #'interpret-slate)
	   (loop for code = (parse-statement token-stream :eof)
	     collect code into result
	     if (eq (peek-token token-stream) :end-statement)
	       do (get-token token-stream)
	     else
	       if (eq (peek-token token-stream) :eof)
	       do (return `(progn ,@(or result '(*primitive-nil*)))))))))))

; Only here for bootstrapping the libraries.
(define-method "fileIn" ((file *traits-file*))
  (invoke-method '|fileIn:| nil *primitive-file* (namestring file)))

; Only here for bootstrapping the libraries.
(define-method "fileIn" ((string *traits-string*))
  (invoke-method '|fileIn:| nil *primitive-file* string))

#+clisp
(progn
(define-method "newOnPort:" ((socket *traits-socket-server*) (port *traits-integer*))
  (socket:socket-server port))
(define-method "newForPeerOf:" ((socket *traits-socket-server*) (stream *traits-socket-client*))
  (socket:socket-server stream))
(define-method "shutdown" ((socket *traits-socket-server*))
  (socket::socket-stream-shutdown socket))
(define-method "close" ((socket *traits-socket-server*))
  (socket:socket-server-close socket))
(define-method "wait" ((socket *traits-socket-server*))
  (socket:socket-wait socket))
(define-method "wait:" ((socket *traits-socket-server*) (delay *traits-number*))
  (socket:socket-wait socket delay))
(define-method "port" ((socket *traits-socket-server*))
  (socket:socket-server-port socket))
(define-method "serverStream" ((socket *traits-socket-server*))
  (socket:socket-accept socket))
(define-method "status" ((socket *traits-socket-server*))
  (socket:socket-status socket))
(define-method "status" ((socket *traits-socket-client*))
  (socket:socket-status socket))
(define-method "host" ((socket *traits-socket-client*))
  (socket::socket-client-peer host))
(define-method "port" ((socket *traits-socket-client*))
  (second (multiple-value-list (socket::socket-client-peer socket))))
(define-method "peerHost" ((socket *traits-socket-client*))
  (socket::socket-client-peer socket))
(define-method "peerPort" ((socket *traits-socket-client*))
  (second (multiple-value-list (socket::socket-client-peer socket))))
(define-method "peerIP" ((socket *traits-socket-client*))
  (socket::socket-client-peer socket t))
(define-method "newOnPort:" ((socket *traits-socket-client*) port)
  (socket:socket-connect port))
(define-method "shutdown" ((socket *traits-socket-client*))
  (socket::socket-client-shutdown socket))
(define-method "close" ((socket *traits-socket-client*))
  (socket::socket-client-shutdown socket)
  (close socket))
; The following methods are intended to write to Sequences of some kind, so
; that Arrays, Strings, and ByteArrays may be used equally.
(define-method "read:from:startingAt:into:" ((socket *traits-socket-client*)
					     n handle start s)
  (read-sequence s socket :start start))

(define-method "write:to:startingAt:from:" ((socket *traits-socket-client*)
					    n handle start s)
  (write-sequence s socket :start start))
(define-method "read:from:startingAt:into:" ((socket *traits-socket-server*)
					     n handle start s)
  (read-sequence s socket :start start))

(define-method "write:to:startingAt:from:" ((socket *traits-socket-server*)
					    n handle start s)
  (write-sequence s socket :start start))
)

(define-method "findOn:" ((selector *traits-symbol*) arguments)
  (let ((arglist (coerce arguments 'list)))
    (prog1
      (dispatch-method selector nil nil arglist)
      (setf (svref arguments 0) (first arglist)))))

(define-method "findOn:after:" ((selector *traits-symbol*) arguments method)
  (let ((arglist (coerce arguments 'list)))
    (prog1
      (dispatch-method selector method nil arglist)
      (setf (svref arguments 0) (first arglist)))))

(define-method "methodNamesAt:" ((object *traits-root*) (index *traits-integer*))
  (method-names-for-position object index))

(define-method "methodsNamed:" ((object *traits-root*)
				(name *traits-symbol*))
  (find-methods object name))

(define-method "methodsNamed:at:" ((object *traits-root*)
				   (name *traits-symbol*)
				   (index *traits-integer*))
  (find-methods object name index))

(define-method "sendTo:" ((selector *traits-symbol*) arguments &optionals optionals)
  (if optionals
      (apply #'invoke-method-with-optionals selector nil optionals (coerce arguments 'list))
      (apply #'invoke-method selector nil (coerce arguments 'list))))

(define-method "sendTo:through:" ((selector *traits-symbol*) arguments dispatchers &optionals optionals)
  (let ((method (dispatch-method selector ignore t dispatchers)))
    (when optionals
      (nconc arguments optionals))
    (apply (the function (get-slot method '|code|)) method arguments)))

(define-method "sendWith:" ((selector *traits-symbol*) x)
  (invoke-method selector nil x))

(define-method "sendWith:with:" ((selector *traits-symbol*) x y)
  (invoke-method selector nil x y))

(define-method "sendWith:with:with:" ((selector *traits-symbol*) x y z)
  (invoke-method selector nil x y z))

(define-method "methodsNamed:atPosition:" ((object *traits-root*) name index)
  (let ((methods (make-array 0 :fill-pointer 0 :adjustable t)))
    (multiple-value-bind (slot exists)
	(gethash name (wm-map-slots (wm-object-map
				     (if (wm-object-p object) object
				       (or (gethash object *literal-objects*)
					   (return #()))))))
      (when exists
	(dolist (role (wm-slot-roles slot))
	  (when (logbitp index (wm-role-positions role))
	    (vector-push-extend (wm-role-method role) methods)))))
    methods))

(define-method "removeFrom:" ((method *traits-method*)
			      (objects *traits-array*))
  (let ((name (get-slot method '|selector|)))
    (loop for each across objects
      for index = 0 then (1+ index)
      for map = (wm-object-map (if (wm-object-p each) each
				  (or (gethash each *literal-objects*)
				      *no-role*)))
      for slot = (gethash name (wm-map-slots map))
      when slot
      do (setf (wm-slot-roles slot)
	       (loop for role in (wm-slot-roles slot)
		 unless (and (logbitp index (wm-role-positions role))
			     (eq method (wm-role-method role)))
		 collect role))))
  method)

(define-method "replaceWith:on:" ((method *traits-method*)
				  (new-method *traits-method*)
				  (objects *traits-array*))
  (let ((name (get-slot method '|selector|)))
    (loop for each across objects
      for index = 0 then (1+ index)
      for map = (wm-object-map (if (wm-object-p each) each
				  (or (gethash each *literal-objects*)
				      *no-role*)))
      do (loop for role in (wm-slot-roles (gethash name (wm-map-slots map)))
	   when (and (logbitp index (wm-role-positions role))
		     (eq method (wm-role-method role)))
	   do (setf (wm-role-method role) new-method))))
  method)

(define-method "identityHash" ((object *traits-root*))
  (sxhash object))

(define-method "raisedTo:" ((x *traits-number*) (y *traits-number*))
  (expt x y))

(define-method "as:" ((i *traits-integer*) (f *traits-float*))
  (coerce i 'float))

(define-method "as:" ((f *traits-float*) (i *traits-integer*))
  (values (floor f)))

(define-method "bitOr:" ((x *traits-integer*) (y *traits-integer*))
  (logior x y))

(define-method "bitXor:" ((x *traits-integer*) (y *traits-integer*))
  (logxor x y))

(define-method "bitAnd:" ((x *traits-integer*) (y *traits-integer*))
  (logand x y))

(define-method "bitShift:" ((x *traits-integer*) (y *traits-integer*))
  (if (> y 0)
      (ash x y)
      (ldb (byte (1- (integer-length x)) (- y)) x)))

(define-method "bitNot" ((x *traits-integer*))
  (lognot x))

(define-method "<<" ((x *traits-integer*) (y *traits-integer*))
  (ash x y))

(define-method ">>" ((x *traits-integer*) (y *traits-integer*))
  (ash x (- y)))

(define-method "+" ((x *traits-integer*) (y *traits-integer*))
  (+ x y))

(define-method "-" ((x *traits-integer*) (y *traits-integer*))
  (- x y))

(define-method "quo:" ((x *traits-integer*) (y *traits-integer*))
  (values (floor x y)))

(define-method "*" ((x *traits-integer*) (y *traits-integer*))
  (* x y))

(define-method "+" ((x *traits-float*) (y *traits-float*))
  (+ x y))

(define-method "-" ((x *traits-float*) (y *traits-float*))
  (- x y))

(define-method "/" ((x *traits-float*) (y *traits-float*))
  (/ x y))

(define-method "*" ((x *traits-float*) (y *traits-float*))
  (* x y))

(define-bool-method "=" ((x *traits-integer*) (y *traits-integer*))
  (= x y))

(define-bool-method ">" ((x *traits-integer*) (y *traits-integer*))
  (> x y))

(define-bool-method "<" ((x *traits-integer*) (y *traits-integer*))
  (< x y))

(define-bool-method ">=" ((x *traits-integer*) (y *traits-integer*))
  (>= x y))

(define-bool-method "<=" ((x *traits-integer*) (y *traits-integer*))
  (<= x y))

(define-bool-method "=" ((x *traits-float*) (y *traits-float*))
  (= x y))

(define-bool-method ">" ((x *traits-float*) (y *traits-float*))
  (> x y))

(define-bool-method "<" ((x *traits-float*) (y *traits-float*))
  (< x y))

(define-bool-method ">=" ((x *traits-float*) (y *traits-float*))
  (>= x y))

(define-bool-method "<=" ((x *traits-float*) (y *traits-float*))
  (<= x y))
 
#+x11 (load "x11")
